#ifdef WIN32
#include <process.h>
#include "win32.h"
#else
#include <sys/time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <net/if.h>
#include <pthread.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>

#include "common.h"
#include "box.h"
#include "console-server.h"
#include "btkcam-client.h"

bisskeys 	bisslist[100];				// BISS & DCW list in dcwbiss.cfg
int		bisscnt=0;

char _logm[4096];
char _flogm[5][16384];
unsigned char _log_mask = 0;

pthread_mutex_t _log_display_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t _log_transaction_mutex = PTHREAD_MUTEX_INITIALIZER;

console_server *remote_console_server = NULL;
btkcam_client *btk[10];

// ---------------------------------- global configuration settings ------------------------------------------
int config_enigma_version = 0;
char binary_file_path[256] = {0};
char config_file_path[256] = {0};

int config_background_execution = 1;
int config_debug_level = 0;
int config_log_to_console = 0;
int config_log_to_file = 0;
char config_logfile_name[256] = {0};
int config_console_logs_broadcast = 0;
int config_console_logs_broadcast_port;
int	operation_mode=7;			// SUM of 	1 - ECM 	2 - BISS 	4 - DCW
int	operation_prio=0;			// 		0 - ECM / DCW   1 - DCW / ECM

// -------------------------------------- Btkcam -------------------
char	Blines[10][255];
int 	Bcnt=0;
// --------------------------------------- global statistics class -------------------------------------------
statistics *stats = NULL;
// -----------------------------------------------------------------------------------------------------------

void config_read_fix_dos_style( char *config_line )
{
	int length = strlen( config_line );
	if( length > 0 )
	{
		if( config_line[length - 1] == (unsigned char)0x0D || config_line[length - 1] == (unsigned char)0x0A)
			config_line[length - 1] = 0;
		
		if( length > 1 )
		{
			if( config_line[length - 2] == (unsigned char)0x0D || config_line[length - 2] == (unsigned char)0x0A)
				config_line[length - 2] = 0;
		}
	}
}

int config_read_settings( void )
{
	FILE *config_file;
	char config_line[1024] = {0};
	int items_read;
	unsigned int temp_int;
	
	
	//for (temp_int=0;temp_int<10;temp_int++) btk[temp_int]=NULL;
	//get the binary file path
	readlink("/proc/self/exe", binary_file_path, sizeof(binary_file_path));
	//printf("Binary path : %s\n", binary_file_path);
	//build the config file name to try
	sprintf(config_line, "/var/keys/btkcam.cfg", binary_file_path);
	
	config_file = fopen(config_line, "rt");
	if (config_file == NULL) 
	{
		_slog(NO_TRANS, _DEBUG_BASIC, 0, "ERROR: Copy Btkcam.cfg into /var/keys folder.", 2, true);
		return 1;
	}
	else { strcpy(config_file_path, config_line); _slog(NO_TRANS, _DEBUG_BASIC, 0, _logmf("Using config file: %s", config_line), 2); }
	
	while( fgets(config_line, 1024, config_file) )
	{
	config_read_fix_dos_style( config_line );
	if( !strncmp("Box: ", config_line, 5) )  sscanf(config_line,"Box: %d",&box::type);
	if( !strncmp("Back: ", config_line, 6) )  sscanf(config_line,"Back: %d",&config_background_execution);
	if( !strncmp("Enigma: ", config_line, 8) )  sscanf(config_line,"Enigma: %d",&config_enigma_version);
	if( !strncmp("Mode: ", config_line, 6) )  sscanf(config_line,"Mode: %d",&operation_mode);
	if( !strncmp("Prio: ", config_line, 6) )  sscanf(config_line,"Prio: %d",&operation_prio);
	if( !strncmp("Dbg_Lvl: ", config_line, 9) ) sscanf(config_line,"Dbg_Lvl: %d",&config_debug_level);
	if( !strncmp("Log_Con: ", config_line, 9) ) sscanf(config_line,"Log_Con: %d",&config_log_to_console); 
	if( !strncmp("Log_File: ", config_line, 10) )  sscanf(config_line,"Log_File: %s %d",config_logfile_name, &config_log_to_file); 
	if( !strncmp("Log_Tcp: ", config_line, 9)) sscanf(config_line,"Log_Tcp: %d %d",&config_console_logs_broadcast_port, &config_console_logs_broadcast); 
	if( !strncmp("B: ", config_line, 3))  memcpy(Blines[Bcnt++],config_line,strlen(config_line));
	}
	fclose( config_file );

	// parse dcw / biss
	sprintf(config_line, "/var/keys/dcwbiss.cfg", binary_file_path);
	
	config_file = fopen(config_line, "rt");
	if (config_file == NULL) return 0;

	while ( fgets ( config_line, 200, config_file ) != NULL )
		   if (strncmp(config_line,"DB: ",4)==0) 
			{
			sscanf(config_line, "DB: %12s %32s",bisslist[bisscnt].pces,bisslist[bisscnt].keys );
			bisscnt++;
			}
	fclose( config_file );
		
	return 0;
}



void config_show_settings( char *settings )
{
	char buffer[1024];
	char line[128];
	int  ixx;
	sprintf(line, "Btkcam Configuration settings:\n\n"); strcpy( buffer, line );
	sprintf(line, "  Box             : %s\n", box::type == 0 ? "Dreambox" : (box::type == 1 ? "Triple Dragon" : (box::type == 2 ? "DGStation Device" : "Invalid"))); strcat( buffer, line );
	sprintf(line, "  Enigma Version  : %s\n", config_enigma_version ? "E2" : "E1"); strcat( buffer, line );
	sprintf(line, "  Background EXE  : %s\n", config_background_execution ? "Yes" : "No"); strcat( buffer, line );
	sprintf(line, "  Debug level     : %i\n", config_debug_level); strcat( buffer, line );
	sprintf(line, "  Log to Console  : %s\n", config_log_to_console ? "Yes" : "No"); strcat( buffer, line );
	sprintf(line, "  Log to file     : %s %s\n", config_log_to_file ? "Yes" : "No",config_logfile_name ); strcat( buffer, line );
	sprintf(line, "  Log to Network  : %s %d\n", config_console_logs_broadcast ? "Yes" : "No",config_console_logs_broadcast_port ); strcat( buffer, line );
	sprintf(line, "  Btkcam  Lines   : %d \n",Bcnt); strcat( buffer, line );
	for (ixx=0;ixx<Bcnt;ixx++) { sprintf(line, "  #%d --> %s \n",ixx,Blines[ixx]); strcat( buffer, line ); }
	sprintf(line, "\n"); strcat( buffer, line );
	if( settings )
		strcpy(settings, buffer);
	else
		printf(buffer);
	
}

void set_debug_level( int level )
{
	config_debug_level = level;
}

void _flog(unsigned int log_transaction, int level, int indent_message, char *message, int indent_data, unsigned char *data, int len, int cols, int newlines_between, int newlines_after, bool show_osd)
{
	int i,j;
	char line[4096];
	
	if( level > config_debug_level ) return;
	
	if( !log_transaction ) _flogm[log_transaction][0] = 0; //if not part of transaction, reset log buffer
	
	//print header
	if( strlen( message ) > 0 )
	{
		for(i=0;i<indent_message;i++) 
		{
			sprintf(line, "  "); strcat( _flogm[log_transaction], line );
		}
		
		sprintf(line, "%s", message); strcat( _flogm[log_transaction], line );
		for(i=0;i<newlines_between;i++) 
		{
			sprintf(line, "\n"); strcat( _flogm[log_transaction], line );
		}
	}
	//print binary data
	if( data != NULL )
	{
		for(i=0;i<len;i++)
		{
			if(cols > 0 )
			{
				if( (i % cols) == 0 )
				{
					if( i > 0 ) 
					{ 
						sprintf(line, "\n"); strcat( _flogm[log_transaction], line );
					}
					if( newlines_between > 0 || strlen( message ) == 0 ) 
						for(j=0;j<indent_data;j++) 
						{
							sprintf(line, "  "); strcat( _flogm[log_transaction], line );
						}
				}
			}
			sprintf(line, "%02X ", data[i]); strcat( _flogm[log_transaction], line );
		}
	}
	for(i=0;i<newlines_after;i++) 
	{
		sprintf(line, "\n"); strcat( _flogm[log_transaction], line );
	}
	if( !log_transaction ) _log_commit_trans( log_transaction, show_osd );
}

void _slog(unsigned int transaction, int level, int indent_message, char *message, int newlines, bool show_osd)
{
	_flog(transaction, level, indent_message, message, 0, NULL, 0, 0, 0, newlines, show_osd);
}

unsigned int _log_begin_trans( void )
{
	pthread_mutex_lock( &_log_transaction_mutex );
	
	for(int i=0;i<4;i++)
	{
		if( (1 << i) & (~_log_mask) )
		{
			_log_mask |= (1 << i);
			_flogm[i + 1][0] = 0; //reset log buffer
			pthread_mutex_unlock( &_log_transaction_mutex );
			return i + 1;
		}
	}
	
	pthread_mutex_unlock( &_log_transaction_mutex );
	return 0;
}

void _log_commit_trans( unsigned int log_trans, bool show_osd )
{
	pthread_mutex_lock( &_log_transaction_mutex );
	pthread_mutex_lock( &_log_display_mutex );

	// -- test log to console
	if( config_log_to_console )
	{
		printf( _flogm[log_trans] );
		if( stats ) stats->webif_console_add_text( _flogm[log_trans] );
	}
	
	// -- test log to file
	if( config_log_to_file )
	{
		FILE *logfile = fopen( config_logfile_name, "a" );
		if( logfile != NULL )
		{
			fprintf(logfile, _flogm[log_trans] );
			fclose( logfile );
		}
	}

	// -- test remote broadcast
	if( remote_console_server )
		{
		remote_console_server->broadcast(_flogm[log_trans]);
		}
	
	// -- test show in osd
	//if( show_osd ) box::on_screen_message(_flogm[log_trans]);
	
	if( log_trans ) _log_mask &= ~(1 << (log_trans - 1));
	
	_flogm[log_trans][0] = 0; //reset log buffer
	
	pthread_mutex_unlock( &_log_display_mutex );
	pthread_mutex_unlock( &_log_transaction_mutex );
}

void _log_abort_trans( unsigned int log_trans )
{
	pthread_mutex_lock( &_log_transaction_mutex );
	_log_mask &= ~(1 << (log_trans - 1));
	_flogm[log_trans][0] = 0; //reset log buffer
	pthread_mutex_unlock( &_log_transaction_mutex );
}

void unix2dos(char *dest, char *source)
{
	int length = strlen(source);
	int dest_index = 0;
	
	for(int i=0;i<length;i++)
	{
		if( source[i] == '\n' )
		{
			dest[dest_index++] = '\r';
			dest[dest_index++] = '\n';
		}	
		else dest[dest_index++] = source[i];
	}
	dest[dest_index] = 0;
}


char *_logmf(const char *format, ...)
{
	va_list arg_list;
	
	va_start( arg_list, format );
	vsprintf( _logm, format, arg_list );
	va_end( arg_list );
	
	return _logm;
}

char *get_time( void )
{
	time_t write_time = time(NULL);
	char *string_time = ctime( &write_time );
	
	int length = strlen(string_time);
	if( length > 0 ) string_time[length - 1] = 0;
	
	return string_time;
}

char *get_time_nf( void )
{
	char hms[9] = {0}, day[3] = {0}, month[4] = {0}, year[5] = {0};
	char *temp_time = get_time();
	char *temp_work = new char[strlen(temp_time) + 1];
	strcpy(temp_work, (const char *)temp_time);
	
	strsub(year, temp_work, 20, 4);
	strsub(month, temp_work, 4, 3);
	if( !strcmp("Jan", month) ) strcpy(month, "01");
	if( !strcmp("Feb", month) ) strcpy(month, "02");
	if( !strcmp("Mar", month) ) strcpy(month, "03");
	if( !strcmp("Apr", month) ) strcpy(month, "04");
	if( !strcmp("May", month) ) strcpy(month, "05");
	if( !strcmp("Jun", month) ) strcpy(month, "06");
	if( !strcmp("Jul", month) ) strcpy(month, "07");
	if( !strcmp("Aug", month) ) strcpy(month, "08");
	if( !strcmp("Sep", month) ) strcpy(month, "09");
	if( !strcmp("Oct", month) ) strcpy(month, "10");
	if( !strcmp("Nov", month) ) strcpy(month, "11");
	if( !strcmp("Dec", month) ) strcpy(month, "12");
	strsub(day, temp_work, 8, 2);
	strsub(hms, temp_work, 11, 8);
	
	sprintf(temp_time, "%04d-%02d-%02d|%s", atoi(year), atoi(month), atoi(day), hms);
	
	delete temp_work;
	return temp_time;
}

void 	keystr_to_key(char *src, unsigned char *cw0, unsigned char *cw1)
{
unsigned char 	keys[16];
unsigned char 	*dst = keys;
unsigned char	*end = keys + sizeof(keys);
unsigned int 	tmpu;
int		i,j;

				
dst = keys;
end = keys + sizeof(keys);
while (dst < end && sscanf(src, "%2x", &tmpu) == 1)
	{
	*dst++ = tmpu;
	src += 2;
	}
for (int j=0; j<8; j++) cw0[j]=keys[j];
for (int j=8; j<16; j++) cw1[j-8]=keys[j];
}


void xor_buffers(unsigned char *dest, unsigned char *src_1, unsigned char *src_2, unsigned int length)
{
	//xor two buffers together.
	while(length--) *dest++ = *src_1++ ^ *src_2++;
}

char *strsub(char *dest, char *src, int start, int length)
{
	strncpy(dest, src + start, length);
	dest[length] = 0;
	return dest;
}

char *strltr(char *dest, char *src)
{
	int i, length = strlen(src);
	for(i=0;i<length;i++) if( src[i] != ' ' && src[i] != '\t' ) break;
	strcpy(dest, &src[i]);
	return dest;
}

char *strrtr(char *dest, char *src)
{
	int i, length = strlen(src);
	strcpy(dest, src);
	if( !length ) return dest;
		
	for(i=length-1;i>=0;i--) if( src[i] != ' ' && src[i] != '\t' ) break;
	dest[i + 1] = 0;
	return dest;
}

char *stratr(char *dest, char *src)
{
	char *temp = new char[strlen(src) + 1];
	strltr(temp, src);
	strrtr(dest, temp);
	delete temp;
	return dest;
}

// --------------------------- key and provider configuration management ------------------------------------


unsigned char *load_file(char *file_path, int *size)
{
	FILE *file;
	unsigned int file_size;
	unsigned char *buffer;
	
	if( (file = fopen(file_path, "rb")) == NULL )
	{
		_slog(NO_TRANS, _DEBUG_BASIC, 2, _logmf("**** ERROR: Could not open %s", file_path), 1);
		return NULL;
	}
	
	fseek( file, 0L, SEEK_END );
	file_size = ftell( file );
	rewind( file );
 	
	if( (buffer = (unsigned char *)malloc( (size_t)file_size )) == NULL )
	{
		_slog(NO_TRANS, _DEBUG_BASIC, 2, _logmf("**** ERROR: Could not allocate memory for %s.", file_path), 1);
		return NULL;
	}
	
	if( fread(buffer, (size_t)1, (size_t)file_size, file) != file_size )
	{
		free( buffer );
		_slog(NO_TRANS, _DEBUG_BASIC, 2, _logmf("**** ERROR: Failed to load complete file %s to memory.", file_path), 1);
		return NULL;
	}
	fclose( file );
	
	*size = file_size;
	return buffer;
}

int fexists(const char * filename)
{
	if (FILE *file = fopen(filename, "r"))
	{
		fclose(file);
		return 1;
	}
	return 0;
}

double microseconds( void )
{
	struct timeval start;
	
	gettimeofday(&start, NULL); 
	return (double)start.tv_sec * (double)1000000 + (double)start.tv_usec;
}

bool memzero(unsigned char *src, unsigned int size)
{
	unsigned char temp[size];
	memset(temp, 0, size);
	return (memcmp(src, temp, size) == 0);
}

bool memequal(unsigned char *src_1, unsigned char *src_2, unsigned int size)
{
	return (memcmp(src_1, src_2, size) == 0);
}




char *file_name_from_path(char *path)
{
	printf("%s\n", path);
	char *current, *last = path;
	while( (current = strstr(last, "/")) ) last = current + 1;
	return last;
}

unsigned int get_local_address( void )
{
	int fd;
	struct ifreq interface;
	struct sockaddr_in socket_address;
	
	fd = socket(PF_INET, SOCK_STREAM, 0);
	strcpy(interface.ifr_name, "eth0");
	ioctl(fd, SIOCGIFADDR, &interface);
	socket_address = *((struct sockaddr_in *)(&(interface.ifr_addr)));
	
	return _ntohl( socket_address.sin_addr.s_addr );
}

char *address_int_to_human_readable(unsigned int address, char *readable)
{
	sprintf(readable, "%d.%d.%d.%d", 
		(address >> 24) & 0xFF, 
		(address >> 16) & 0xFF, 
		(address >>  8) & 0xFF, 
		(address >>  0) & 0xFF);
}

int xtoi(char *hex)
{
	int value, items_read;
	items_read = sscanf(hex, "%X", &value);
	
	if( items_read > 0 )
		return value;
	else
		return 0;
}

void dump(unsigned char *buffer, int length)
{
	int i;
	
	printf("0x0=");
	for (i = 0;i < length;i++)
	{
		if (i % 16 == 0) printf("\n%04X=", i);
		printf("%02X", buffer[i]);
	}
	printf("\n");
}

// -------------------------------------------------------------------------------
// ------------------- these are manually implemented as a -----------------------
// ------------- workaround for images that do not include them ------------------

unsigned int _htonl(unsigned int hostlong)
{
	unsigned int netlong = 0;
	((unsigned char *)(&netlong))[0] = (hostlong >> 24) && 0xFF;
	((unsigned char *)(&netlong))[1] = (hostlong >> 16) && 0xFF;
	((unsigned char *)(&netlong))[2] = (hostlong >>  8) && 0xFF;
	((unsigned char *)(&netlong))[3] = (hostlong >>  0) && 0xFF;
	return netlong;
}

unsigned short int _htons(unsigned short int hostshort)
{
	unsigned short int netshort;
	((unsigned char *)(&netshort))[0] = (hostshort >> 8) & 0xFF;
	((unsigned char *)(&netshort))[1] = (hostshort >> 0) & 0xFF;
	return netshort;
}

unsigned int _ntohl(unsigned int netlong)
{
	return ((((unsigned char *)(&netlong))[0]) << 24) | ((((unsigned char *)(&netlong))[1]) << 16) | 
		((((unsigned char *)(&netlong))[2]) << 8) | ((((unsigned char *)(&netlong))[3]) << 0);
}

unsigned short int _ntohs(unsigned short int netshort)
{
	return ((((unsigned char *)(&netshort))[0]) << 8) | ((((unsigned char *)(&netshort))[1]) << 0);
}
